/*
 * SOI 2015, Creativity Task "Bazaar"
 * Server by benjamin@soi.ch
 */

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Scanner;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.TimeUnit;

@SuppressWarnings("serial")
class ProtocolViolatedException extends Exception {
	public ProtocolViolatedException(String message) {
		super(message);
	}
}

@SuppressWarnings("serial")
class BadUserInputException extends Exception {
	public BadUserInputException(String message) {
		super(message);
	}
}

class Program {
	public String id;
	private String command;
	private Process process;
	private PrintWriter programOut;
	private BufferedReader programIn, programErr;

	private ArrayBlockingQueue<Integer> programInQ;
	private PutProgramInToQueueThread putProgramInToQueueThread;
	private StdErrHandlerThread stdErrHandlerThread;

	private final int LONG_TIMEOUT = 1750;
	private final int SHORT_TIMEOUT = 750;
	private final int MAX_STDERR_LINES = 1000;

	public Program(String cmd, String id) {
		this.command = cmd;
		this.id = id;
	}

	public void start() throws IOException {
		process = Runtime.getRuntime().exec(command);

		programOut = new PrintWriter(process.getOutputStream(), true);
		programIn = new BufferedReader(new InputStreamReader(
				process.getInputStream()));
		programErr = new BufferedReader(new InputStreamReader(
				process.getErrorStream()));

		programInQ = new ArrayBlockingQueue<Integer>(1, true);
		putProgramInToQueueThread = new PutProgramInToQueueThread(programInQ,
				programIn);
		putProgramInToQueueThread.start();
		stdErrHandlerThread = new StdErrHandlerThread(id, programErr);
		stdErrHandlerThread.start();
	}

	public void feedStartData(int N, int A, int D, int G) {
		programOut.println(N + " " + A + " " + D + " " + G);
	}

	public void feedOpponentBids(int ignoreNum, int[] bids) {
		StringBuilder msg = new StringBuilder();
		for (int i = 0; i < bids.length; i++) {
			if (i == ignoreNum) {
				continue;
			}
			msg.append(bids[i] + " ");
		}
		msg.setLength(msg.length() - 1);
		programOut.println(msg);
	}

	public Integer getMove() throws ProtocolViolatedException {
		try {
			Integer res;
			res = programInQ.poll(SHORT_TIMEOUT, TimeUnit.MILLISECONDS);
			if (res == null) {
				// no move was available within timeout
				if (putProgramInToQueueThread.isAlive()) {
					// program is still running, try again
					res = programInQ.poll(LONG_TIMEOUT, TimeUnit.MILLISECONDS);
					if (res == null) {
						// Programm is blocked or in endless loop
						throw new ProtocolViolatedException(id
								+ " could not print a line within "
								+ (SHORT_TIMEOUT + LONG_TIMEOUT) + "ms");
					} else {
						Log.progBehavior(id, "needed between " + SHORT_TIMEOUT
								+ "ms and " + (SHORT_TIMEOUT + LONG_TIMEOUT)
								+ "ms to print a line, that's quite long!");
					}
				} else {
					// Program crashed
					throw new ProtocolViolatedException(id
							+ " terminated instead ot printing a line");
				}
			}
			return res;
		} catch (InterruptedException e) {
			throw new RuntimeException(e);
		}
	}

	volatile boolean terminating = false;

	public void terminate() throws IOException {
		synchronized (this) {
			if (terminating)
				return;
			terminating = true;
		}

		putProgramInToQueueThread.terminating = true;
		putProgramInToQueueThread.interrupt();
		stdErrHandlerThread.terminating = true;
		// no need to interrupt stdErrHandlerThread, it will terminate because
		// stderr is closed

		Thread terminator = new Thread() {
			@Override
			public void run() {
				try {
					Log.progBehavior(id,
							"exited with status " + process.waitFor());
				} catch (InterruptedException e) {
					process.destroy();
					Log.progBehavior(id,
							"did not properly exit and was therefore killed");
				}
			}
		};
		terminator.start();

		try {
			terminator.join(50);
			if (terminator.isAlive()) {
				terminator.interrupt();
			}
		} catch (InterruptedException e) {
			throw new RuntimeException(e);
		}

		programOut.close();
		programIn.close();
		programErr.close();
	}

	class PutProgramInToQueueThread extends Thread {
		private ArrayBlockingQueue<Integer> queue;
		private BufferedReader reader;
		public volatile boolean terminating = false;

		public PutProgramInToQueueThread(ArrayBlockingQueue<Integer> queue,
				BufferedReader reader) {
			this.queue = queue;
			this.reader = reader;
		}

		@Override
		public void run() {
			try {
				while (true) {
					String line = reader.readLine();
					if (line == null) {
						// program has terminated
						break;
					} else {
						queue.put(Integer.parseInt(line));
					}
				}
			} catch (InterruptedException e) {
				if (!terminating)
					e.printStackTrace();
			} catch (IOException e) {
				if (!terminating)
					e.printStackTrace();
			}
		}
	}

	class StdErrHandlerThread extends Thread {
		private String id;
		private BufferedReader reader;
		public volatile boolean terminating = false;

		public StdErrHandlerThread(String id, BufferedReader reader) {
			this.id = id;
			this.reader = reader;
		}

		@Override
		public void run() {
			try {
				for (int i = 0; i < MAX_STDERR_LINES; i++) {
					String line = reader.readLine();
					if (line == null) {
						// program has terminated
						return;
					} else {
						Log.progStdErr(id, line);
					}
				}
			} catch (IOException e) {
				if (!terminating)
					e.printStackTrace();
			}
		}
	}
}

class Game {
	private int N, A, D, G;
	private String[] playerCommands, playerIds;
	private Program[] players;

	public Game(int A, int D, int G, String[] playerCommands, String[] playerIds) {
		this.N = playerCommands.length;
		this.A = A;
		this.D = D;
		this.G = G;
		this.playerCommands = playerCommands.clone();
		this.playerIds = playerIds.clone();
	}

	public double[] playGame() {
		players = new Program[N];
		double[] diamonds = new double[N];

		for (int n = 0; n < N; n++) {
			players[n] = new Program(playerCommands[n], playerIds[n]);
			diamonds[n] = 0;
		}

		try {
			Log.tech("Gonna start all bots");
			for (int n = 0; n < N; n++) {
				players[n].start();
				players[n].feedStartData(N, A, D, G);
			}
			Log.tech("All bots started");

			Log.game("Start game with N = " + N + ", A = " + A + ", D = " + D
					+ ", G = " + G);

			StringBuilder msg = new StringBuilder();
			for (int i = 0; i < players.length; i++) {
				msg.append(players[i].id + ", ");
			}
			msg.setLength(msg.length() - 2);
			Log.game(N + " players: " + msg);

			double[] res;
			int nextpercentage = 5;
			for (int a = 0; a < A; a++) {
				res = playAuction();
				for (int n = 0; n < N; n++) {
					diamonds[n] += res[n];
				}

				while (a * 100 > A * nextpercentage) {
					Log.tech(nextpercentage + "%");
					nextpercentage += 5;
				}
			}
		} catch (ProtocolViolatedException e) {
			Log.tech("Protocol violation: " + e.getMessage());

			for (int n = 0; n < N; n++) {
				diamonds[n] = 0;
			}
		} catch (IOException e) {
			e.printStackTrace();
		}

		Log.tech("Game finished, gonna kill the bots");
		for (int n = 0; n < N; n++) {
			try {
				players[n].terminate();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
		Log.tech("Terminated all bots. Mission accomplished");

		Log.game("game finished");
		for (int n = 0; n < N; n++) {
			Log.game(players[n].id + ": "
					+ (Math.round(diamonds[n] * 100.0) / 100.0));
		}

		return diamonds;
	}

	private double[] playAuction() throws ProtocolViolatedException {
		Log.game("Start new auction");

		double diamonds[] = new double[N];
		int gold[] = new int[N];
		int bids[] = new int[N];
		for (int n = 0; n < N; n++) {
			gold[n] = G;
			diamonds[n] = 0;
		}

		for (int d = 0; d < D; d++) {
			int maxBid = 0;
			for (int n = 0; n < N; n++) {
				try {
					bids[n] = players[n].getMove();
					gold[n] -= bids[n];
					Log.game(players[n].id + " bids " + bids[n]);
					if (gold[n] < 0) {
						throw new ProtocolViolatedException(players[n].id
								+ " spent more gold than allowed");
					}
				} catch (ProtocolViolatedException e) {
					Log.game("Game cannot be played to end because "
							+ players[n].id + " did not respect protocol");
					throw e;
				}
				maxBid = Math.max(maxBid, bids[n]);
			}
			int maxBidCount = 0;
			for (int n = 0; n < N; n++) {
				if (bids[n] == maxBid) {
					maxBidCount++;
				}
			}
			for (int n = 0; n < N; n++) {
				if (bids[n] == maxBid) {
					diamonds[n] += 1.0 / maxBidCount;
					Log.game(players[n].id + " receives "
							+ (maxBidCount == 1 ? "1" : "1/" + maxBidCount)
							+ " diamond");
				}

				players[n].feedOpponentBids(n, bids);
			}
		}

		Log.game("auction finished");
		for (int n = 0; n < N; n++) {
			Log.game(players[n].id + ": "
					+ (Math.round(diamonds[n] * 100.0) / 100.0));
		}

		return diamonds;
	}
}

class Log {
	private static PrintWriter gameLogWriter;

	synchronized static void game(String s) {
		if (gameLogWriter == null) {
			System.out.println(s);
		} else {
			gameLogWriter.println(s);
		}
	}

	static void setGameLogPath(String path) {
		if (gameLogWriter != null) {
			gameLogWriter.close();
		}
		if (path == null) {
			gameLogWriter = null;
		} else {
			try {
				gameLogWriter = new PrintWriter(path);
			} catch (FileNotFoundException e) {
				gameLogWriter = null;
			}
		}
	}

	synchronized static void tech(String s) {
		System.err.println(s);
	}

	synchronized static void progStdErr(String id, String s) {
		System.err.println("[stderr of " + id + "] " + s);
	}

	synchronized static void progBehavior(String id, String s) {
		System.err.println(id + " " + s);
	}
}

public class Server {
	public static void main(String[] args) {
		new Server(args);
	}

	public Server(String[] args) {
		if (args.length == 0) {
			printHelp();
			return;
		}

		int N, A, D, G;
		String[] commands;
		String[] ids;
		N = A = D = G = 0;
		ids = new String[0];
		commands = new String[0];
		try {
			int cmdoffset = -1, idoffset = -1;
			String cmdfile = null, idfile = null;
			for (int i = 0; i < args.length; i++) {
				if (args[i].equals("-game")) {
					A = Integer.parseInt(args[i + 1]);
					D = Integer.parseInt(args[i + 2]);
					G = Integer.parseInt(args[i + 3]);
					check(A > 0, "A > 0 is violated");
					check(D > 0, "D > 0 is violated");
					check(G > 0, "G > 0 is violated");
					i += 3;
				} else if (args[i].equals("-player")) {
					N = Integer.parseInt(args[i + 1]);
					check(N > 1, "N > 1 is violated");
					cmdoffset = i + 2;
					i += N + 1;
				} else if (args[i].equals("-playerfile")) {
					cmdfile = args[i + 1];
					i++;
				} else if (args[i].equals("-id")) {
					idoffset = i + 1;
					i += N;
				} else if (args[i].equals("-idfile")) {
					idfile = args[i + 1];
					i++;
				} else if (args[i].equals("-log")) {
					Log.setGameLogPath(args[i + 1]);
					i++;
				}
			}
			commands = new String[N];
			if (cmdoffset > 0) {
				for (int n = 0; n < N; n++) {
					commands[n] = args[cmdoffset + n];
				}
			}
			if (cmdfile != null) {
				commands = readAllLines(cmdfile);
				N = commands.length;
				check(N > 1, "N > 1 is violated");
			}

			ids = new String[N];
			for (int n = 0; n < N; n++) {
				ids[n] = "Player" + (n + 1);
			}
			if (idoffset > 0) {
				for (int n = 0; n < N; n++) {
					ids[n] = args[idoffset + n];
				}
			}
			if (idfile != null) {
				ids = readAllLines(idfile);
				check(ids.length == N, "bad id file: wrong number of ids");
			}
			check(A > 0, "missing game info");
			check(N > 0 && commands[0] != null, "missing player info");
		} catch (BadUserInputException e) {
			System.out.println(e.getMessage());
			printHelp();
			return;
		} catch (Exception e) {
			System.out.println("Couldn't parse arguments");
			printHelp();
			// throw e;
			return;
		}

		Game game = new Game(A, D, G, commands, ids);
		game.playGame();

		Log.setGameLogPath(null);
	}

	private void check(boolean cond, String msg) throws BadUserInputException {
		if (!cond) {
			throw new BadUserInputException(msg);
		}
	}

	private String[] readAllLines(String file) throws BadUserInputException {
		try {
			ArrayList<String> lines = new ArrayList<String>();

			Scanner sc = new Scanner(new File(file));
			while (sc.hasNextLine()) {
				String line = sc.nextLine();
				if (!(line.trim().isEmpty() || line.trim().startsWith("#"))) {
					lines.add(line);
				}
			}

			sc.close();

			String[] res = new String[lines.size()];
			lines.toArray(res);

			return res;
		} catch (Exception e) {
			throw new BadUserInputException("bad player file: "
					+ e.getMessage());
		}
	}

	private void printHelp() {
		System.out.println("Usage: java Server [-game A D G] "
				+ "[-player N playerCommand1 ... playerCommandN] "
				+ "[-playerfile file] " + "[-id playerID1 ... playerIDN] "
				+ "[-idfile file] " + "[-log file]");
		System.out.println("You have to specify at least gameinfos (-game)\n"
				+ "and players (-player or -playerfile).");
		System.out.println("In a playerfile/idfile list one item per line.");
	}
}
